"""Hyyp API constants."""
from enum import Enum

PUSH_DELAY = 30
DEFAULT_TIMEOUT = 25
MAX_RETRIES = 3
GCF_SENDER_ID = 87969245803
GCM_APP_ID = "1:87969245803:android:100fe8e62b328f50"
REQUEST_HEADER = {
    "User-Agent": "okhttp/3.12.1",
}  # Standard android header.

IMEI_SEED = "15241971"
STD_PARAMS = {
    "imei": None,  # Alphabet soup starts at 0
    "appVersionCode": "428",
    "_appVersionCode": "428",
    "deviceOS": "12.0",
    "deviceName": "Python API",
    "pkg": "za.co.adt.securehome.android",
    "_appVersionName": "3.5.19",
    "token": None,
    "userId": None,
}  # Standard request parameters.

DEBUG_CLIENT_STRING = {
    "client_string" : None
}

# Rpc to name mapping. Used in push notifications.
RpcCodes = {
    "202": "IMEI unknown to SMART platform. Registration failed.",
    "203": "Command failed",
    "204": "VALUE is invalid",
    "205": "User code invalid",
    "206": "Unit busy",
    "207": "Command was older than abort command time",
    "208": "Communications timeout to peripheral attached to Hub",
    "209": "Invalid PID in topic",
    "210": "Serial code mismatch",
}

# EventCategory to name mapping.
# Used in notifications.
EventCategory = {"1": "Emergency", "2": "User", "3": "Trouble", "4": "Information"}


class HyypPkg(Enum):
    """Supported hyyp skins/rebranding."""

    ADT_SECURE_HOME = "za.co.adt.securehome.android"
    IDS_HYYP_GENERIC = "com.hyyp247.home"


# EventNumber to name mapping.
EventNumber = {
    "50": "AC clock",
    "13": "AC fail",
    "24": "AC Restore",
    "3": "Alarm cancel",
    "7": "Armed with bypassed zone",
    "43": "Auto arm cancel",
    "0": "Away arm",
    "17": "Battery low",
    "28": "Battery low restore",
    "20": "Box tamper",
    "31": "Box tamper restore",
    "22": "Bus comms fail",
    "33": "Bus comms restore",
    "23": "Bus device battery low",
    "34": "Bus device battery restore",
    "21": "Bus device tamper",
    "32": "Bus device tamper restore",
    "14": "Comms fail",
    "25": "Comm restore",
    "53": "Crossed zone alarm",
    "89": "Cross zone trigger",
    "51": "Crystal oscillator clock",
    "35": "Dedicated panic",
    "2": "Disarm",
    "54": "Disarm from stay",
    "37": "Download access",
    "66": "DTMF login",
    "38": "Duress",
    "18": "12V fuse fail",
    "29": "12V fuse restore",
    "19": "Engineer reset",
    "30": "Engineer reset restore",
    "88": "Entry delay",
    "85": "Exit delay",
    "67": "Failed IDSwift login",
    "90": "Forced door",
    "52": "Installer code changed",
    "45": "Installer mode",
    "40": "Keypad fire",
    "42": "Keypad lockout",
    "41": "Keypad medical",
    "39": "Keypad panic",
    "79": "MPS AC fail",
    "82": "MPS AC restore",
    "80": "MPS batt low",
    "83": "MPS Batt restore",
    "81": "MPS Fuse fail",
    "84": "MPS Fuse restore",
    "48": "Panel defaulted",
    "26": "Phone line restore",
    "15": "Phone line tamper",
    "47": "Power up",
    "92": "Remote arming request",
    "93": "Remote panic",
    "49": "Reserved",
    "72": "RF Detector batt restore",
    "68": "RF Detector low batt",
    "69": "RF Detector supervision loss",
    "73": "RF Detector supervision restore",
    "70": "RF Jam",
    "74": "RF Jam restore",
    "71": "Rf RSSI low",
    "75": "RF RSSI restore",
    "27": "Siren restore",
    "16": "Siren tamper",
    "1": "Stay arm",
    "77": "Stay zone report",
    "230": "System trouble",
    "231": "System trouble restore",
    "91": "TAG arming request",
    "94": "Tamper alarm",
    "100": "User trigger",
    "101": "User panic",
    "36": "Test report",
    "78": "Time stamp",
    "65": "User bypassed zones",
    "46": "User codes defaulted",
    "44": "User code changed",
    "76": "User menu accessed",
    "87": "User unbypassed zones",
    "86": "Voice login lockout",
    "8": "Force armed",
    "6": "Zone restore",
    "11": "Zone shutdown",
    "12": "Zone shutdown restore",
    "9": "Zone tamper",
    "10": "Zone tamper restore",
    "5": "Zone violation alarm",
}



"""Receive GCM/FCM messages from google."""

from base64 import b64encode, urlsafe_b64decode, urlsafe_b64encode
from binascii import hexlify
import json
import logging
import os
import os.path
import select
import socket
import ssl
import struct
import time
from urllib.parse import urlencode
from urllib.request import Request, urlopen
import threading as thread
import requests
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from google.protobuf.json_format import MessageToDict
import http_ece
from oscrypto.asymmetric import generate_pair



from .common_tools import ClientTools
from .android_checkin_pb2 import AndroidCheckinProto, ChromeBuildProto
from .checkin_pb2 import AndroidCheckinRequest, AndroidCheckinResponse
from .constants import GCF_SENDER_ID, GCM_APP_ID
from .mcs_pb2 import (
    Close,
    DataMessageStanza,
    HeartbeatAck,
    HeartbeatPing,
    IqStanza,
    LoginRequest,
    LoginResponse,
    StreamErrorStanza,
)

_LOGGER = logging.getLogger(__name__)

SERVER_KEY = (
    b"\x04\x33\x94\xf7\xdf\xa1\xeb\xb1\xdc\x03\xa2\x5e\x15\x71\xdb\x48\xd3"
    + b"\x2e\xed\xed\xb2\x34\xdb\xb7\x47\x3a\x0c\x8f\xc4\xcc\xe1\x6f\x3c"
    + b"\x8c\x84\xdf\xab\xb6\x66\x3e\xf2\x0c\xd4\x8b\xfe\xe3\xf9\x76\x2f"
    + b"\x14\x1c\x63\x08\x6a\x6f\x2d\xb1\x1a\x95\xb0\xce\x37\xc0\x9c\x6e"
)


APP_VER = "4.2.8"

REGISTER_URL = "https://android.clients.google.com/c2dm/register3"
CHECKIN_URL = "https://android.clients.google.com/checkin"
FCM_SUBSCRIBE = "https://fcm.googleapis.com/fcm/connect/subscribe"
FCM_ENDPOINT = "https://fcm.googleapis.com/fcm/send"

FCM_AUTH_URL = "https://firebaseinstallations.googleapis.com/v1/projects/hyyp-49b11/installations/"
FCM_REGISTRATION_URL = "https://fcmtoken.googleapis.com/register"
FCM_V1_REGISTRATION_URL = "https://fcmregistrations.googleapis.com/v1/"

GOOGLE_MTALK_ENDPOINT = "mtalk.google.com"
READ_TIMEOUT_SECS = 60 * 60
MIN_RESET_INTERVAL_SECS = 60 * 5
MAX_SILENT_INTERVAL_SECS = 60 * 15

GOOGLE_FCM_PUBLIC_APIKEY = "AIzaSyDH5H6kfGQEWm7FQSYfWYy8OAHPq__5Y6s"

MCS_VERSION = 41
PACKET_BY_TAG = [
    HeartbeatPing,
    HeartbeatAck,
    LoginRequest,
    LoginResponse,
    Close,
    "MessageStanza",
    "PresenceStanza",
    IqStanza,
    DataMessageStanza,
    "BatchPresenceStanza",
    StreamErrorStanza,
    "HttpRequest",
    "HttpResponse",
    "BindAccountRequest",
    "BindAccountResponse",
    "TalkMetadata",
]




class FCMRegistration:
    
    
    def __init__(
        self,
    ) -> None:

        self.blank = '123'

    def __do_request(self, req, retries=5):
        for _ in range(retries):
            try:
                resp = urlopen(req)
                resp_data = resp.read()
                resp.close()
                _LOGGER.debug(resp_data)                
                return resp_data
            except Exception as err:
                _LOGGER.debug("error during request", exc_info=err)
                time.sleep(1)
        return None


    def gcm_check_in(self, androidId=None, securityToken=None, **kwargs):
        """
        perform check-in request

        androidId, securityToken can be provided if we already did the initial
        check-in

        returns dict with androidId, securityToken and more
        """
        chrome = ChromeBuildProto()
        chrome.platform = 3
        chrome.chrome_version = "63.0.3234.0"
        chrome.channel = 1

        checkin = AndroidCheckinProto()
        checkin.type = 3
        checkin.chrome_build.CopyFrom(chrome)  # pylint: disable=maybe-no-member

        payload = AndroidCheckinRequest()
        payload.user_serial_number = 0
        payload.checkin.CopyFrom(checkin)  # pylint: disable=maybe-no-member
        payload.version = 3
        if androidId:
            payload.id = int(androidId)
        if securityToken:
            payload.security_token = int(securityToken)

        _LOGGER.debug(payload)
        req = Request(
            url=CHECKIN_URL,
            headers={"Content-Type": "application/x-protobuf"},
            data=payload.SerializeToString(),
        )
        resp_data = self.__do_request(req)
        resp = AndroidCheckinResponse()
        resp.ParseFromString(resp_data)
        _LOGGER.debug(resp)
        resp = MessageToDict(resp)        
        return resp


    def urlsafe_base64(self, data):
        """
        base64-encodes data with -_ instead of +/ and removes all = padding.
        also strips newlines

        returns a string
        """
        res = urlsafe_b64encode(data).replace(b"=", b"")
        return res.replace(b"\n", b"").decode("ascii")



# start of new FCM functions


    def fcm_get_initial_auth_data(self, app_id, credentials = None, retries = 5):
        """
        get the google firebase auth key for the app. app_id is provided, second half being random (see def register())    

        returns the firebase auth key and the supplied (generated) app_id
        """
       
        data =  {
                "appId":app_id,
                "sdkVersion":"i:8.10.0"
            }
        
        headers = {
                "x-goog-api-key" : GOOGLE_FCM_PUBLIC_APIKEY,        
        }
       
        _LOGGER.debug(data)
        #error checks need to be added
        req = requests.post(url=FCM_AUTH_URL, json=data, headers=headers)
        fcm_auth_data = req.json()
        fcm_auth_data["appid"] = app_id
        return fcm_auth_data



    def fcm_get_token(self, fcm_auth_data):

        
        check_in = self.gcm_check_in() 
        firebase_installation_auth = fcm_auth_data["authToken"]["token"]
        appid = fcm_auth_data["fid"]
        gmp_app_id = fcm_auth_data["appid"]
        androidid = check_in["androidId"]
        security_token = check_in["securityToken"]
        versioninfo = check_in["versionInfo"]
      
        public, private = generate_pair("ec", curve=str("secp256r1"))
        keys = {
            "public": self.urlsafe_base64(public.asn1.dump()[26:]),
            "private": self.urlsafe_base64(private.asn1.dump()),
            "secret": self.urlsafe_base64(os.urandom(16)),
        }
                
        headers = {
                "x-goog-firebase-installations-auth":firebase_installation_auth,
                "authorization":"AidLogin " + androidid + ":" + security_token,
                "info":versioninfo,
                "app":"com.hyyp247.home"
        }
         
        data = {
            "device":androidid,
            "app":"com.hyyp247.home",
            "app_ver":APP_VER,
            "X-cliv":"fiid-8.10.0",
            "sender":GCF_SENDER_ID,
            "X-subtype":GCF_SENDER_ID,
            "appid":appid,
            "gmp_app_id":gmp_app_id,
            
            }       
          
        data = urlencode(data)
        req = requests.post(url=FCM_REGISTRATION_URL + "?" + data, headers=headers)
        fcm_token = req.text
        res = {"fcm": fcm_token}
        res.update(check_in)
        res.update(fcm_auth_data)
        res.update(keys)
        return res

    def fcm_subscribe(self, credentials):

        firebase_installation_auth = credentials["auth_info"]["googleAuthToken"]
        androidid = credentials["gcm"]["androidId"]
        fid = credentials["auth_info"]["fid"]
        gmp_app_id = credentials["gcm"]["appId"]
        security_token = credentials["gcm"]["securityToken"]
        fcm_token = credentials["fcm"]["token"]
        refresh_token = credentials["fcm"]["refreshToken"]

                 
        headers = {
                "x-goog-firebase-installations-auth":firebase_installation_auth,
                "authorization":"AidLogin " + androidid + ":" + security_token,
        }
      
        data = {
            "device":androidid,
            "app":"com.hyyp247.home",
            "app_ver":APP_VER,
            "X-cliv":"fiid-8.10.0",
            "sender":GCF_SENDER_ID,
            "X-subtype":GCF_SENDER_ID,
            "refreshToken":refresh_token,
            "appid":fid,
            "gmp_app_id":gmp_app_id,
            "endpoint": "{}/{}".format(FCM_ENDPOINT, fcm_token),
            "p256dh": credentials["keys"]["public"],
            "auth": credentials["keys"]["secret"],           
            }       
      
       
        data = urlencode(data)
        req = requests.post(url=FCM_REGISTRATION_URL + "?" + data, headers=headers)
        return req




#this isn't being used. I tested this but it doesn't seem to be working correctly
# I am keeping the code to not lose the work I put in

    def fcm_register_with_v1_api(self, credentials, retries = 5):  
        print("using v1 api")
        firebase_installation_auth = credentials["auth_info"]["googleAuthToken"]
        androidid = credentials["gcm"]["androidId"]
        security_token = credentials["gcm"]["securityToken"]
        fcm_token = credentials["fcm"]["token"]
 
 
        headers = {
                "x-goog-api-key" : GOOGLE_FCM_PUBLIC_APIKEY, 
                "x-goog-firebase-installations-auth":firebase_installation_auth,
                "authorization":"AidLogin " + androidid + ":" + security_token,
        }
 
         
        data = {"web":{
                "endpoint": "{}/{}".format(FCM_ENDPOINT, fcm_token),
                "p256dh": credentials["keys"]["public"],
                "auth": credentials["keys"]["secret"],
                }
        }    
           
        
        url_full = FCM_V1_REGISTRATION_URL + "projects/" + str(GCF_SENDER_ID) + "/registrations"
        _LOGGER.debug(data)
        print("---")
        resp_data = requests.post(url=url_full, json=data, headers=headers)
        return resp_data

        
    
    


    def build_credetials_json(self, all_info):
        
        
        appid = all_info["appid"]
        android_id = all_info["androidId"]
        security_token = all_info["securityToken"]
        google_auth_token = all_info["authToken"]["token"]
        refresh_token = all_info["refreshToken"]
        version_info = all_info["versionInfo"]
        name = all_info["name"]
        fid = all_info["fid"]
        fcmtoken = all_info["fcm"].replace("token=","")
        publickey = all_info["public"]
        privatekey = all_info["private"]
        secretkey = all_info["secret"]
        
        
        
        processed_credentials = {
                                    "gcm":{
                                        "appId":appid,
                                        "androidId":android_id,
                                        "securityToken":security_token,
                                        "versionInfo":version_info,            
                                    },
                                    "fcm":{
                                        "token":fcmtoken,
                                        "refreshToken":refresh_token,                                      
                                    },
                                    "auth_info":{
                                        "versionInfo":version_info,
                                        "googleAuthToken":google_auth_token,
                                        "name":name,
                                        "fid":fid,                                     
                                    },
                                    "keys":{
                                        "public":publickey,
                                        "private":privatekey,
                                        "secret":secretkey, 
                                    }
            
        }
      

        return processed_credentials
        
        
        


    def register(self):
        """register gcm and fcm tokens for sender_id"""
        #app_id = "1:87969245803:ios:" + os.urandom(8).hex()
        app_id = GCM_APP_ID
        subscription = self.fcm_get_initial_auth_data(app_id=app_id)
        credentials_unprocessed = self.fcm_get_token(fcm_auth_data=subscription)
        credentials = self.build_credetials_json(credentials_unprocessed)
        self.fcm_subscribe(credentials=credentials)
        return credentials







#end of new fcm functions





# -------------------------------------------------------------------------


class FCMListener:

    
    def __init__(
        self,
    ) -> None:
        self.fcm_registration = FCMRegistration()
        self.received_persistent_ids = []
        self.time_of_last_reset = 0
        self.time_of_last_receive = time.time()
        self.current_ping_thread = 0
        self.listen_for_data_thread = 0
        self.awaiting_ack = False
        self.ids_callback = None


    def __read(self, sock, size):
        _LOGGER.debug("Started reading")
        _LOGGER.debug("Size: " + str(size))
        _count = 0
        buf = b""
        while len(buf) < size:
            _count += 1
            buf += sock.recv(size - len(buf))
            if _count > 10:
                _LOGGER.debug("Escaped infite loop")
                sock.close()
                return buf
        _LOGGER.debug("Finished reading")
        return buf
        

    # protobuf variable length integers are encoded in base 128
    # each byte contains 7 bits of the integer and the msb is set if there's
    # more. pretty simple to implement


    def __read_varint32(self, value):
        res = 0
        shift = 0
        while True:
            (b,) = struct.unpack("B", self.__read(value, 1))
            res |= (b & 0x7F) << shift
            if (b & 0x80) == 0:
                break
            shift += 7
        return res


    def __encode_varint32(self, value):
        res = bytearray([])
        while value != 0:
            b = value & 0x7F
            value >>= 7
            if value != 0:
                b |= 0x80
            res.append(b)
        return bytes(res)


    def __send(self, google_socket, data):
        header = bytearray([MCS_VERSION, PACKET_BY_TAG.index(type(data))])
        _LOGGER.debug(data)
        payload = data.SerializeToString()
        buf = bytes(header) + self.__encode_varint32(len(payload)) + payload
        _LOGGER.debug(hexlify(buf))
        total = 0
        while total < len(buf):
            sent = google_socket.send(buf[total:])
            if sent == 0:
                raise RuntimeError("socket connection broken")
            total += sent


    def __recv(self, data, first=False):

        try:
            readable, _, _ = select.select(
                [
                    data,
                ],
                [],
                [],
                READ_TIMEOUT_SECS,
            )
            if len(readable) == 0:
                _LOGGER.debug("Select read timeout")
                return None

        except select.error:
            _LOGGER.debug("Select error")
            return None

        _LOGGER.debug("Data available to read")

        if first:
            version, tag = struct.unpack("BB", self.__read(data, 2))
            _LOGGER.debug("version %s", version)
            if version < MCS_VERSION and version != 38:
                raise RuntimeError("protocol version {} unsupported".format(version))
        else:
            (tag,) = struct.unpack("B", self.__read(data, 1))
        _LOGGER.debug("tag %s (%s)", tag, PACKET_BY_TAG[tag])
        size = self.__read_varint32(data)
        _LOGGER.debug("size %s", size)
        self.time_of_last_receive = time.time()
        if size >= 0:
            buf = self.__read(data, size)
            _LOGGER.debug(hexlify(buf))
            packet = PACKET_BY_TAG[tag]
            payload = packet()
            payload.ParseFromString(buf)
            _LOGGER.debug(payload)
            return payload
        return None


    def __app_data_by_key(self, data, key, blow_shit_up=True):
        for item in data.app_data:
            if item.key == key:
                return item.value
        if blow_shit_up:
            raise RuntimeError("couldn't find in app_data {}".format(key))
        return None


    def __open(self):

        context = ssl.create_default_context()
        google_socket = socket.create_connection((GOOGLE_MTALK_ENDPOINT, 5228))
        google_socket = context.wrap_socket(
            google_socket, server_hostname=GOOGLE_MTALK_ENDPOINT
        )
        _LOGGER.debug("connected to ssl socket")
        return google_socket


    def __login(self, credentials, persistent_ids):
        google_socket = self.__open()
        self.fcm_registration.gcm_check_in(**credentials["gcm"])
        req = LoginRequest()
        req.adaptive_heartbeat = False
        req.auth_service = 2
        req.auth_token = credentials["gcm"]["securityToken"]
        req.id = "chrome-63.0.3234.0"
        req.domain = "mcs.android.com"
        req.device_id = "android-%x" % int(credentials["gcm"]["androidId"])
        req.network_type = 1
        req.resource = credentials["gcm"]["androidId"]
        req.user = credentials["gcm"]["androidId"]
        req.use_rmq2 = True
        req.setting.add(name="new_vc", value="1")  # pylint: disable=maybe-no-member
        req.received_persistent_id.extend(persistent_ids)  # pylint: disable=maybe-no-member
        self.__send(google_socket, req)
        login_response = self.__recv(google_socket, first=True)
        _LOGGER.debug("Received login response: %s", login_response)
        self.time_of_last_receive = time.time() + 60
        thread.Thread(target=self.__ping_scheduler, args=(google_socket,
                                                          credentials,
                                                          persistent_ids)).start()

        return google_socket


    def __reset(self, google_socket, credentials, persistent_ids):
        now = time.time()
        if now - self.time_of_last_reset < MIN_RESET_INTERVAL_SECS:
            raise Exception("Too many connection reset attempts.")
        self.time_of_last_reset = now
        _LOGGER.debug("Reestablishing connection")
        try:
            google_socket.shutdown(2)
            google_socket.close()
        except OSError as err:
            _LOGGER.debug("Unable to close connection %f", err)
        return self.__login(credentials, persistent_ids)


    def _close_socket(self, google_socket):
        try:
            google_socket.shutdown(2)
            google_socket.close()
        except OSError as err:
            _LOGGER.debug("Unable to close connection %f", err)
            
    def _restart_push_receiver(self, google_socket):
        self.current_ping_thread += 1
        self.awaiting_ack = False
        self.time_of_last_receive = time.time() + 60
        self.listen_for_data_thread += 1
        self._close_socket(google_socket)
        _LOGGER.debug("RESTARTING PUSH RECEIVER")
        self.ids_callback("restart_push_receiver")
        


    def __ping_scheduler(self, google_socket, credentials, persistent_ids):
        self.current_ping_thread += 1
        if self.current_ping_thread > 1000:
            self.current_ping_thread = 1
        mythread = self.current_ping_thread
        while mythread == self.current_ping_thread:
            if self.awaiting_ack:
                _LOGGER.debug(str(mythread) + ": Ping Timeout resetting")
                self._restart_push_receiver(google_socket=google_socket)
                break
            if time.time() - self.time_of_last_receive > MAX_SILENT_INTERVAL_SECS:
                _LOGGER.debug(str(mythread) + ": Sending PING now==========================")
                self.awaiting_ack = True
                try:
                    self.__send_ping(google_socket=google_socket)
                except:
                    _LOGGER.debug("Error with ping send %f")
                    self._restart_push_receiver(google_socket=google_socket)
                    break
            if not self._internet_connectivity():
                self.awaiting_ack = True
            time.sleep(60)
        _LOGGER.debug("Closing PING thread : " + str(mythread))

                
                    
    def __send_ping(self, google_socket):
        header = bytearray([0, 0])
        buf = bytes(header)
        total = 0
        while total < len(buf):
            sent = google_socket.send(buf[total:])
            if sent == 0:
                raise RuntimeError("socket connection broken")
            total += sent
     
     
    def __handle_ping(self, google_socket):
        header = bytearray([0, 2])
        buf = bytes(header)
        total = 0
        while total < len(buf):
            sent = google_socket.send(buf[total:])
            if sent == 0:
                raise RuntimeError("socket connection broken")
            total += sent
  
     
    def __listen(self, credentials, on_notify_callback, ids_callback, received_persistent_ids=None, obj=None):
        received_persistent_ids = []
        if received_persistent_ids is None:
            received_persistent_ids = []   
        persistent_ids = received_persistent_ids
        self.ids_callback = ids_callback
        self.fcm_registration.fcm_subscribe(credentials=credentials)
        google_socket = self.__login(credentials, persistent_ids)
        self.listen_for_data_thread += 1
        mythread = self.listen_for_data_thread
        while mythread == self.listen_for_data_thread:
            try:
                data = self.__recv(google_socket)
                if isinstance(data, DataMessageStanza):
                    msg_id = self.__handle_data_message(data, credentials, on_notify_callback, obj)
                    persistent_ids.append(msg_id)
                elif isinstance(data, HeartbeatPing):
                    self.__handle_ping(google_socket)
                elif isinstance(data, HeartbeatAck):
                    self.awaiting_ack = False
                elif data is None or isinstance(data, Close):
                    self.listen_for_data_thread += 1
                else:
                    _LOGGER.debug("Unexpected message type %s", type(data))
            except ConnectionResetError:
                _LOGGER.debug("Connection Reset: Reconnecting")
                self.listen_for_data_thread += 1
            except:
                _LOGGER.debug("Other Listener Error")
                self.listen_for_data_thread += 1
        _LOGGER.debug("Closing main thread" + str(mythread))
        self._close_socket(google_socket=google_socket)    
                


    def __handle_data_message(self, data, credentials, callback, obj):
        load_der_private_key = serialization.load_der_private_key

        crypto_key = self.__app_data_by_key(
            data, "crypto-key", blow_shit_up=False
        )  # Can be None
        if crypto_key:
            crypto_key = crypto_key[3:]  # strip dh=

        salt = self.__app_data_by_key(data, "encryption", blow_shit_up=False)  # Can be None
        if salt:
            salt = salt[5:]  # strip salt=

        crypto_key = urlsafe_b64decode(crypto_key.encode("ascii"))
        salt = urlsafe_b64decode(salt.encode("ascii"))
        der_data = credentials["keys"]["private"]
        der_data = urlsafe_b64decode(der_data.encode("ascii") + b"========")
        secret = credentials["keys"]["secret"]
        secret = urlsafe_b64decode(secret.encode("ascii") + b"========")
        privkey = load_der_private_key(der_data, password=None, backend=default_backend())
        decrypted = http_ece.decrypt(
            data.raw_data,
            salt=salt,
            private_key=privkey,
            dh=crypto_key,
            version="aesgcm",
            auth_secret=secret,
        )
        _LOGGER.debug("Received data message %s: %s", data.persistent_id, decrypted)
        callback(obj, json.loads(decrypted.decode("utf-8")), data)
        return data.persistent_id
   
   
    def debug_threaded_runner(self, callback, credentials = None):
        thread.Thread(target=self.runner, args=(callback,
                                                    credentials)).start()
        
    def runner(self, callback, credentials = None):
        """Registers a token and waits for notifications"""
        #_LOGGER.setLevel(logging.DEBUG)
        ids_callback = callback
        
        if credentials is None:
            credentials = self.fcm_registration.register()
            _credentials = {"credentials" : credentials}
            callback(_credentials) #doesn't do anything for now. Using a different method currently

        def on_notification(obj, notification, data_message):
            idstr = data_message.persistent_id
            if idstr in self.received_persistent_ids:
                _LOGGER.warning("Found pers ID, skipping")
                return
            self.received_persistent_ids.append(idstr)
            _LOGGER.warning("ID STR: " + idstr) #test
            _LOGGER.warning("^^ ID STR") #test
            _LOGGER.warning("ID STR array: " + self.received_persistent_ids) #test
            _LOGGER.warning("^^ ID STR array") #test
            _notification = {"notification" : notification}
            callback(_notification)
            _LOGGER.debug(_notification)   
         
        self.__listen(credentials=credentials, on_notify_callback=on_notification, ids_callback=ids_callback, received_persistent_ids=self.received_persistent_ids)
        
    def _internet_connectivity(self):
        return ClientTools().internet_connectivity()


"""Hyyp Api Exceptions."""


class HyypApiError(Exception):
    """ADT Secure Home api exception."""


class InvalidURL(HyypApiError):
    """Invalid url exception."""


class HTTPError(HyypApiError):
    """Invalid host exception."""



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mcs.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\tmcs.proto\x12\tmcs_proto\"S\n\rHeartbeatPing\x12\x11\n\tstream_id\x18\x01 \x01(\x05\x12\x1f\n\x17last_stream_id_received\x18\x02 \x01(\x05\x12\x0e\n\x06status\x18\x03 \x01(\x03\"R\n\x0cHeartbeatAck\x12\x11\n\tstream_id\x18\x01 \x01(\x05\x12\x1f\n\x17last_stream_id_received\x18\x02 \x01(\x05\x12\x0e\n\x06status\x18\x03 \x01(\x03\"a\n\tErrorInfo\x12\x0c\n\x04\x63ode\x18\x01 \x02(\x05\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\x12\'\n\textension\x18\x04 \x01(\x0b\x32\x14.mcs_proto.Extension\"&\n\x07Setting\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\r\n\x05value\x18\x02 \x02(\t\"A\n\rHeartbeatStat\x12\n\n\x02ip\x18\x01 \x02(\t\x12\x0f\n\x07timeout\x18\x02 \x02(\x08\x12\x13\n\x0binterval_ms\x18\x03 \x02(\x05\"G\n\x0fHeartbeatConfig\x12\x13\n\x0bupload_stat\x18\x01 \x01(\x08\x12\n\n\x02ip\x18\x02 \x01(\t\x12\x13\n\x0binterval_ms\x18\x03 \x01(\x05\"\xdb\x02\n\x0b\x43lientEvent\x12)\n\x04type\x18\x01 \x01(\x0e\x32\x1b.mcs_proto.ClientEvent.Type\x12\x1f\n\x17number_discarded_events\x18\x64 \x01(\r\x12\x15\n\x0cnetwork_type\x18\xc8\x01 \x01(\x05\x12#\n\x1atime_connection_started_ms\x18\xca\x01 \x01(\x04\x12!\n\x18time_connection_ended_ms\x18\xcb\x01 \x01(\x04\x12\x13\n\nerror_code\x18\xcc\x01 \x01(\x05\x12\'\n\x1etime_connection_established_ms\x18\xac\x02 \x01(\x04\"[\n\x04Type\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x14\n\x10\x44ISCARDED_EVENTS\x10\x01\x12\x15\n\x11\x46\x41ILED_CONNECTION\x10\x02\x12\x19\n\x15SUCCESSFUL_CONNECTION\x10\x03J\x06\x08\xc9\x01\x10\xca\x01\"\xff\x03\n\x0cLoginRequest\x12\n\n\x02id\x18\x01 \x02(\t\x12\x0e\n\x06\x64omain\x18\x02 \x02(\t\x12\x0c\n\x04user\x18\x03 \x02(\t\x12\x10\n\x08resource\x18\x04 \x02(\t\x12\x12\n\nauth_token\x18\x05 \x02(\t\x12\x11\n\tdevice_id\x18\x06 \x01(\t\x12\x13\n\x0blast_rmq_id\x18\x07 \x01(\x03\x12#\n\x07setting\x18\x08 \x03(\x0b\x32\x12.mcs_proto.Setting\x12\x1e\n\x16received_persistent_id\x18\n \x03(\t\x12\x1a\n\x12\x61\x64\x61ptive_heartbeat\x18\x0c \x01(\x08\x12\x30\n\x0eheartbeat_stat\x18\r \x01(\x0b\x32\x18.mcs_proto.HeartbeatStat\x12\x10\n\x08use_rmq2\x18\x0e \x01(\x08\x12\x12\n\naccount_id\x18\x0f \x01(\x03\x12\x39\n\x0c\x61uth_service\x18\x10 \x01(\x0e\x32#.mcs_proto.LoginRequest.AuthService\x12\x14\n\x0cnetwork_type\x18\x11 \x01(\x05\x12\x0e\n\x06status\x18\x12 \x01(\x03\x12,\n\x0c\x63lient_event\x18\x16 \x03(\x0b\x32\x16.mcs_proto.ClientEvent\"\x1d\n\x0b\x41uthService\x12\x0e\n\nANDROID_ID\x10\x02J\x04\x08\x13\x10\x14J\x04\x08\x14\x10\x15J\x04\x08\x15\x10\x16\"\xf6\x01\n\rLoginResponse\x12\n\n\x02id\x18\x01 \x02(\t\x12\x0b\n\x03jid\x18\x02 \x01(\t\x12#\n\x05\x65rror\x18\x03 \x01(\x0b\x32\x14.mcs_proto.ErrorInfo\x12#\n\x07setting\x18\x04 \x03(\x0b\x32\x12.mcs_proto.Setting\x12\x11\n\tstream_id\x18\x05 \x01(\x05\x12\x1f\n\x17last_stream_id_received\x18\x06 \x01(\x05\x12\x34\n\x10heartbeat_config\x18\x07 \x01(\x0b\x32\x1a.mcs_proto.HeartbeatConfig\x12\x18\n\x10server_timestamp\x18\x08 \x01(\x03\"/\n\x11StreamErrorStanza\x12\x0c\n\x04type\x18\x01 \x02(\t\x12\x0c\n\x04text\x18\x02 \x01(\t\"\x07\n\x05\x43lose\"%\n\tExtension\x12\n\n\x02id\x18\x01 \x02(\x05\x12\x0c\n\x04\x64\x61ta\x18\x02 \x02(\x0c\"\xdd\x02\n\x08IqStanza\x12\x0e\n\x06rmq_id\x18\x01 \x01(\x03\x12(\n\x04type\x18\x02 \x02(\x0e\x32\x1a.mcs_proto.IqStanza.IqType\x12\n\n\x02id\x18\x03 \x02(\t\x12\x0c\n\x04\x66rom\x18\x04 \x01(\t\x12\n\n\x02to\x18\x05 \x01(\t\x12#\n\x05\x65rror\x18\x06 \x01(\x0b\x32\x14.mcs_proto.ErrorInfo\x12\'\n\textension\x18\x07 \x01(\x0b\x32\x14.mcs_proto.Extension\x12\x15\n\rpersistent_id\x18\x08 \x01(\t\x12\x11\n\tstream_id\x18\t \x01(\x05\x12\x1f\n\x17last_stream_id_received\x18\n \x01(\x05\x12\x12\n\naccount_id\x18\x0b \x01(\x03\x12\x0e\n\x06status\x18\x0c \x01(\x03\"4\n\x06IqType\x12\x07\n\x03GET\x10\x00\x12\x07\n\x03SET\x10\x01\x12\n\n\x06RESULT\x10\x02\x12\x0c\n\x08IQ_ERROR\x10\x03\"%\n\x07\x41ppData\x12\x0b\n\x03key\x18\x01 \x02(\t\x12\r\n\x05value\x18\x02 \x02(\t\"\xf4\x02\n\x11\x44\x61taMessageStanza\x12\n\n\x02id\x18\x02 \x01(\t\x12\x0c\n\x04\x66rom\x18\x03 \x02(\t\x12\n\n\x02to\x18\x04 \x01(\t\x12\x10\n\x08\x63\x61tegory\x18\x05 \x02(\t\x12\r\n\x05token\x18\x06 \x01(\t\x12$\n\x08\x61pp_data\x18\x07 \x03(\x0b\x32\x12.mcs_proto.AppData\x12\x1b\n\x13\x66rom_trusted_server\x18\x08 \x01(\x08\x12\x15\n\rpersistent_id\x18\t \x01(\t\x12\x11\n\tstream_id\x18\n \x01(\x05\x12\x1f\n\x17last_stream_id_received\x18\x0b \x01(\x05\x12\x0e\n\x06reg_id\x18\r \x01(\t\x12\x16\n\x0e\x64\x65vice_user_id\x18\x10 \x01(\x03\x12\x0b\n\x03ttl\x18\x11 \x01(\x05\x12\x0c\n\x04sent\x18\x12 \x01(\x03\x12\x0e\n\x06queued\x18\x13 \x01(\x05\x12\x0e\n\x06status\x18\x14 \x01(\x03\x12\x10\n\x08raw_data\x18\x15 \x01(\x0c\x12\x15\n\rimmediate_ack\x18\x18 \x01(\x08\"\x0b\n\tStreamAck\"\x1a\n\x0cSelectiveAck\x12\n\n\x02id\x18\x01 \x03(\tB\x02H\x03')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mcs_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'H\003'
  _globals['_HEARTBEATPING']._serialized_start=24
  _globals['_HEARTBEATPING']._serialized_end=107
  _globals['_HEARTBEATACK']._serialized_start=109
  _globals['_HEARTBEATACK']._serialized_end=191
  _globals['_ERRORINFO']._serialized_start=193
  _globals['_ERRORINFO']._serialized_end=290
  _globals['_SETTING']._serialized_start=292
  _globals['_SETTING']._serialized_end=330
  _globals['_HEARTBEATSTAT']._serialized_start=332
  _globals['_HEARTBEATSTAT']._serialized_end=397
  _globals['_HEARTBEATCONFIG']._serialized_start=399
  _globals['_HEARTBEATCONFIG']._serialized_end=470
  _globals['_CLIENTEVENT']._serialized_start=473
  _globals['_CLIENTEVENT']._serialized_end=820
  _globals['_CLIENTEVENT_TYPE']._serialized_start=721
  _globals['_CLIENTEVENT_TYPE']._serialized_end=812
  _globals['_LOGINREQUEST']._serialized_start=823
  _globals['_LOGINREQUEST']._serialized_end=1334
  _globals['_LOGINREQUEST_AUTHSERVICE']._serialized_start=1287
  _globals['_LOGINREQUEST_AUTHSERVICE']._serialized_end=1316
  _globals['_LOGINRESPONSE']._serialized_start=1337
  _globals['_LOGINRESPONSE']._serialized_end=1583
  _globals['_STREAMERRORSTANZA']._serialized_start=1585
  _globals['_STREAMERRORSTANZA']._serialized_end=1632
  _globals['_CLOSE']._serialized_start=1634
  _globals['_CLOSE']._serialized_end=1641
  _globals['_EXTENSION']._serialized_start=1643
  _globals['_EXTENSION']._serialized_end=1680
  _globals['_IQSTANZA']._serialized_start=1683
  _globals['_IQSTANZA']._serialized_end=2032
  _globals['_IQSTANZA_IQTYPE']._serialized_start=1980
  _globals['_IQSTANZA_IQTYPE']._serialized_end=2032
  _globals['_APPDATA']._serialized_start=2034
  _globals['_APPDATA']._serialized_end=2071
  _globals['_DATAMESSAGESTANZA']._serialized_start=2074
  _globals['_DATAMESSAGESTANZA']._serialized_end=2446
  _globals['_STREAMACK']._serialized_start=2448
  _globals['_STREAMACK']._serialized_end=2459
  _globals['_SELECTIVEACK']._serialized_start=2461
  _globals['_SELECTIVEACK']._serialized_end=2487
# @@protoc_insertion_point(module_scope)



"""init hyyp api exceptions."""
from .alarm_info import HyypAlarmInfos
from .client import HyypClient
from .constants import GCF_SENDER_ID, HyypPkg
from .exceptions import HTTPError, HyypApiError, InvalidURL
#from .push_receiver import run_example

__all__ = [
    "HyypClient",
    "InvalidURL",
    "HTTPError",
    "HyypApiError",
    "HyypPkg",
    "GCF_SENDER_ID",
    "HyypAlarmInfos",
]



"""
Generates random IMEI numbers.

The user specifies the 8-digit TAC and up to 4-digits of the serial number.
The user also specifies the number of random IMEIs to generate.
"""
import random
from .constants import STD_PARAMS


class ImeiGenerator():

    def __init__(
        self,

    ) -> None:
        """Initialize the client object."""

    def checksum(self, number, alphabet='0123456789'):
        """
        Calculate the Luhn checksum over the provided number.

        The checksum is returned as an int.
        Valid numbers should have a checksum of 0.
        """
        n = len(alphabet)
        number = tuple(alphabet.index(i)
                    for i in reversed(str(number)))
        return (sum(number[::2]) +
                sum(sum(divmod(i * 2, n))
                    for i in number[1::2])) % n


    def calc_check_digit(self, number, alphabet='0123456789'):
        """Calculate the extra digit."""
        check_digit = self.checksum(number + alphabet[0])
        return alphabet[-check_digit]


    def generate_imei(self, start = None):
        """Ask for the base IMEI, how many to generate, then generate them."""
        # Loop until the first 8-12 digits have been received & are valid
        if start is None:
            start = STD_PARAMS["imei"]
        # Loop until we know how many random numbers to generate

        # IMEIs will be generated based on the first 8 digits (TAC; the number
        #   used to identify the model) and the next 2-6 digits (partial serial #).
        #   The final, 15th digit, is the Luhn algorithm check digit.
        # Generate and print random IMEI numbers

        imei = start

            # Randomly compute the remaining serial number digits
        while len(imei) < 14:
            imei += str(random.randint(0, 9))

            # Calculate the check digit with the Luhn algorithm
            imei += self.calc_check_digit(imei)

        return imei


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='pyhyypapihawkmod',
    version="1.5.0b6",
    license='Apache Software License 2.0',
    author='hawky358',
    author_email='hawky358@users.github.com',
    description='IDS Hyyp/ADT Secure Home API',
    long_description="API for accessing IDS Hyyp. This is used by ADT Home Connect and possibly others.",
    url='https://github.com/hawky358/pyHyypApi',
    packages=setuptools.find_packages(),
    setup_requires=[
        'requests',
        'setuptools'
    ],
    install_requires=[
        'requests',
        'pandas',
        'oscrypto',
        'protobuf',
        'http-ece',
        'appdirs'
    ],
    python_requires = '>=3.6'
)



"""Hyyp Client API."""
from __future__ import annotations

import logging
from typing import Any

import requests
import threading as thread
import time

from .alarm_info import HyypAlarmInfos
from .push_receiver import FCMListener, FCMRegistration
from .constants import DEFAULT_TIMEOUT, REQUEST_HEADER, STD_PARAMS, DEBUG_CLIENT_STRING, PUSH_DELAY, IMEI_SEED, HyypPkg
from .exceptions import HTTPError, HyypApiError, InvalidURL
from .imei import ImeiGenerator
from .common_tools import ClientTools

_LOGGER = logging.getLogger(__name__)

BASE_URL = "ids.trintel.co.za/Inhep-Impl-1.0-SNAPSHOT/"
API_ENDPOINT_LOGIN = "/auth/login"
API_ENDPOINT_CHECK_APP_VERSION = "/auth/checkAppVersion"
API_ENDPOINT_GET_SITE_NOTIFICATIONS = "/device/getSiteNotifications"
API_ENDPOINT_SYNC_INFO = "/device/getSyncInfo"
API_ENDPOINT_STATE_INFO = "/device/getStateInfo"
API_ENDPOINT_NOTIFICATION_SUBSCRIPTIONS = "/device/getNotificationSubscriptions"
API_ENDPOINT_GET_USER_PREFERANCES = "/user/getUserPreferences"
API_ENDPOINT_SET_USER_PREFERANCE = "/user/setUserPreference"
API_ENDPOINT_SECURITY_COMPANIES = "/security-companies/list"
API_ENDPOINT_STORE_GCM_REGISTRATION_ID = "/user/storeGcmRegistrationId"
API_ENDPOINT_ARM_SITE = "/device/armSite"
API_ENDPOINT_TRIGGER_ALARM = "/device/triggerAlarm"
API_ENDPOINT_SET_ZONE_BYPASS = "/device/bypass"
API_ENDPOINT_GET_CAMERA_BY_PARTITION = "/device/getCameraByPartition"
API_ENDPOINT_UPDATE_SUB_USER = "/user/updateSubUser"
API_ENDPOINT_SET_NOTIFICATION_SUBSCRIPTIONS = "/user/setNotificationSubscriptionsNew"
API_ENDPOINT_TRIGGER_AUTOMATION = "/device/trigger"
API_ENDPOINT_GET_ZONE_STATE_INFO = "/device/getZoneStateInfo"
REQUEST_PUSH_TIMEOUT = 1.5
FAILURE_CAUSE_STRING = "failure_cause"
class HyypClient:
    """Initialize api client object."""

    def __init__(
        self,
        email: str | None = None,
        password: str | None = None,
        pkg: str = HyypPkg.ADT_SECURE_HOME.value,
        timeout: int = DEFAULT_TIMEOUT,
        token: str | None = None,
        userid: int | None = None,
    ) -> None:
        """Initialize the client object."""
        self.alarminfos = HyypAlarmInfos(self)
        self.fcm_listener = FCMListener()
        self.fcm_register = FCMRegistration()
        self.tools = ClientTools()
        self._email = email
        self._password = password
        self._session = requests.session()
        self._session.headers.update(REQUEST_HEADER)
        STD_PARAMS["pkg"] = pkg
        STD_PARAMS["token"] = token
        STD_PARAMS["userId"] = userid
        STD_PARAMS["imei"] = self.generate_imei()
        self.fcm_credentials = None
        self._timeout = timeout
        self.time_to_push = PUSH_DELAY
        self.forced_refresh = False
        self.current_status = None
        self.generic_callback_to_hass = None
    
        # print(self.fcm_credentials)
        
    def login(self) -> Any:
        """Login to ADT Secure Home API."""
           
        _params = STD_PARAMS.copy()
        _params["email"] = self._email
        _params["password"] = self._password
        try:
            req = self._session.get(
                "https://" + BASE_URL + API_ENDPOINT_LOGIN,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )

            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err

        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(f"Login error: {_json_result['error']}")

        STD_PARAMS["token"] = _json_result["token"]
        STD_PARAMS["userId"] = _json_result["user"]["id"]

        DEBUG_CLIENT_STRING["client_string"] = _json_result
        return _json_result

    def generate_imei(self):
        imei = ImeiGenerator().generate_imei(IMEI_SEED)
        return imei

    def check_app_version(self) -> Any:
        """Check App version via API."""

        _params = STD_PARAMS.copy()
        _params["clientImei"] = STD_PARAMS["imei"]

        try:
            req = self._session.get(
                "https://" + BASE_URL + API_ENDPOINT_CHECK_APP_VERSION,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )

            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err

        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(
                f"Error checking app version from api: {_json_result['error']}"
            )

        return _json_result

    def alarm_info_push_timer(self, callback, onetime = False):
        SLEEP_DELAY = 0.1
        if onetime:
            time.sleep(REQUEST_PUSH_TIMEOUT)
            alarminfo = self.load_alarm_infos()
            callback(alarminfo)
            return
        while 1:
            if self.forced_refresh and self.time_to_push > REQUEST_PUSH_TIMEOUT:
                self.time_to_push = REQUEST_PUSH_TIMEOUT
            while self.time_to_push > 0:
                time.sleep(SLEEP_DELAY)
                self.time_to_push -= SLEEP_DELAY
            alarminfo = self.load_alarm_infos()
            callback(alarminfo)
            self.forced_refresh = False
            self.time_to_push = PUSH_DELAY

    def request_alarm_info_push_to_hass(self):
        self.forced_refresh = True
        self.time_to_push = REQUEST_PUSH_TIMEOUT
        
    def initialize_alarm_info_push_timer(self, callback, onetime = False):
        thread.Thread(target=self.alarm_info_push_timer,
                      kwargs={"callback" : callback, "onetime" : onetime}).start()


    def load_alarm_infos(self) -> dict[Any, Any]:
        """Get alarm infos formatted for hass infos."""
        forced = self.forced_refresh
        self.current_status = self.alarminfos.status(forced=forced)
        return self.current_status

    def initialize_fcm_notification_listener(self, callback, restart = False):
        thread.Thread(target=self.fcm_notification_thread,
                      kwargs={"callback" : callback,
                              "restart" : restart,
                              }).start()
   
   
    def register_generic_callback_to_hass(self, callback):
       self.generic_callback_to_hass = callback

    def fcm_notification_thread(self, callback, restart):
        if not restart:
            if not self.tools.internet_connectivity():
                return
        if self.fcm_credentials is None:
            self.fcm_credentials = self.get_initial_fcm_credentials()
        if self.fcm_credentials is None:
            _LOGGER.warning("No FCM credentials available, disabling notifications")
            return
        if "fcm" not in self.fcm_credentials:
            _LOGGER.warning("No FCM credentials available, disabling notifications")
            return
        gcm_address = self.fcm_credentials["fcm"]["token"]  
        if restart:
            time.sleep(10)
        if not self.tools.internet_connectivity():
            while not self.tools.internet_connectivity():
                time.sleep(60) 
            callback("restart_push_receiver")
            return
        time.sleep(2)

        
        retry_count = 0
        while self.store_gcm_registrationid(gcm_id=gcm_address) == 0:
            time.sleep(30)
            retry_count += 1
            if retry_count >= 2:
                callback("restart_push_receiver")
                return
        time.sleep(60)
        self.fcm_listener.runner(callback=callback,
                                 credentials=self.fcm_credentials)


    def get_initial_fcm_credentials(self):
        # if not self.tools.internet_connectivity():
            # return False
        res = self.fcm_register.register()
        _LOGGER.warning("From Get init ^^^")
        _LOGGER.warning(res)
        return res
        return self.fcm_register.register()
        ## add some sort of validation

    def get_debug_infos(self) -> dict[Any, Any]:
        """Get alarm infos formatted for hass infos."""
        return HyypAlarmInfos(self).get_debug_info()

    def site_notifications(
        self, site_id: int, timestamp: int | None = None, json_key: int | None = None
    ) -> Any:
        """Get site notifications from API."""

        _params: dict[str, Any] = STD_PARAMS.copy()
        _params["siteId"] = site_id
        _params["timestamp"] = timestamp

        try:
            req = self._session.get(
                "https://" + BASE_URL + API_ENDPOINT_GET_SITE_NOTIFICATIONS,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )

            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err

        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(
                f"Error getting site notifications from api: {_json_result['error']}"
            )

        if json_key is None or not _json_result["listSiteNotifications"][str(site_id)]:
            return _json_result["listSiteNotifications"][str(site_id)]

        return _json_result["listSiteNotifications"][str(site_id)][json_key]

    def set_notification_subscriptions(
        self,
        trouble_notifications: bool = True,
        emergency_notifications: bool = True,
        user_notifications: bool = True,
        information_notifications: bool = True,
        test_report_notifications: bool = False,
    ) -> Any:
        """Enable or disable app notifications."""

        _params: dict[str, Any] = STD_PARAMS.copy()
        del _params["imei"]
        _params["mobileImei"] = STD_PARAMS["imei"]
        _params["troubleNotifications"] = trouble_notifications
        _params["emergencyNotifications"] = emergency_notifications
        _params["userNotifications"] = user_notifications
        _params["informationNotifications"] = information_notifications
        _params["testReportNotifications"] = test_report_notifications

        try:
            req = self._session.post(
                "https://" + BASE_URL + API_ENDPOINT_SET_NOTIFICATION_SUBSCRIPTIONS,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )

            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err

        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(
                f"Error getting site notifications from api: {_json_result['error']}"
            )

        return _json_result

    def get_camera_by_partition(
        self, partition_id: int, json_key: str | None = None
    ) -> Any:
        """Get cameras, bypassed zones and zone ids by partition from API."""

        _params: dict[str, Any] = STD_PARAMS.copy()
        _params["partitionId"] = partition_id

        try:
            req = self._session.get(
                "https://" + BASE_URL + API_ENDPOINT_GET_CAMERA_BY_PARTITION,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )

            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err

        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(
                f"Error getting partition cameras from api: {_json_result['error']}"
            )
        if json_key is None:
            return _json_result

        return _json_result[json_key]



    def get_zone_state_info(self, site_id: int, json_key: str | None = None) -> Any:
        """Get state info from API. Returns armed, bypassed partition ids."""

        _params: dict[str, Any] = STD_PARAMS.copy()
        _params["siteId"] = site_id
        
        try:
            req = self._session.get(
                "https://" + BASE_URL + API_ENDPOINT_GET_ZONE_STATE_INFO,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )
            req.raise_for_status()
        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err    
        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            _LOGGER.warning(f"Error getting zone state info from api: {_json_result['error']}")
            #raise HyypApiError(
            #    f"Error getting zone state info from api: {_json_result['error']}"
            #)

        if json_key is None:
            return _json_result

        return _json_result[json_key]


    def get_sync_info(self, json_key: str | None = None) -> Any:
        """Get user, site, partition and users info from API."""

        _params = STD_PARAMS.copy()
        _params["_appPkg"] = _params["pkg"]
        if self.fcm_credentials is not None and "fcm" in self.fcm_credentials:
            _params["gcmId"] = self.fcm_credentials["fcm"]["token"]
            _LOGGER.warning("^^^ In sync info")
            _LOGGER.warning(_params["gcmId"])
        try:
            req = self._session.get(
                "https://" + BASE_URL + API_ENDPOINT_SYNC_INFO,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )

            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err

        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(
                f"Error getting sync info from api: {_json_result['error']}"
            )

        if json_key is None:
            return _json_result

        return _json_result[json_key]

    def get_state_info(self, json_key: str | None = None) -> Any:
        """Get state info from API. Returns armed, bypassed partition ids."""

        _params = STD_PARAMS

        try:
            req = self._session.get(
                "https://" + BASE_URL + API_ENDPOINT_STATE_INFO,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )

            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err

        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(
                f"Error getting state info from api: {_json_result['error']}"
            )

        if json_key is None:
            return _json_result

        return _json_result[json_key]

    def get_notification_subscriptions(self, json_key: str | None = None) -> Any:
        """Get notification subscriptions from API."""

        _params = STD_PARAMS

        try:
            req = self._session.get(
                "https://" + BASE_URL + API_ENDPOINT_NOTIFICATION_SUBSCRIPTIONS,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )

            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err

        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(
                f"Error getting notification subscriptions: {_json_result['error']}"
            )

        if json_key is None:
            return _json_result

        return _json_result[json_key]

    def get_user_preferences(
        self, user_id: int, site_id: int | None = None, json_key: str | None = None
    ) -> Any:
        """Get user preferences from API."""

        _params: dict[str, Any] = STD_PARAMS.copy()
        _params["userId"] = user_id
        _params["siteId"] = site_id

        try:
            req = self._session.get(
                "https://" + BASE_URL + API_ENDPOINT_GET_USER_PREFERANCES,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )

            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err

        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(
                f"Error getting user preferences: {_json_result['error']}"
            )

        if json_key is None:
            return _json_result

        return _json_result[json_key]

    def get_security_companies(self, json_key: str | None = None) -> Any:
        """Get security companies from API."""

        _params = STD_PARAMS

        try:
            req = self._session.get(
                "https://" + BASE_URL + API_ENDPOINT_SECURITY_COMPANIES,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )

            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err

        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(
                f"Failed to get security companies: {_json_result['error']}"
            )

        if json_key is None:
            return _json_result

        return _json_result[json_key]

    def store_gcm_registrationid(self, gcm_id = None) -> Any:
        """Store gcmid."""
        _params = STD_PARAMS.copy()
        _params["gcmId"] = gcm_id
        del _params["imei"]
        _params["clientImei"] = STD_PARAMS["imei"]
        try:
            req = self._session.post(
                "https://" + BASE_URL + API_ENDPOINT_STORE_GCM_REGISTRATION_ID,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )
        
            req.raise_for_status()
         
        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err
        
        except:
            _LOGGER.debug("GCM Registration Error")
            return 0
        
        

        try:
            _json_result = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(f"Storing gcm id failed with: {_json_result['error']}")
        return _json_result

    def set_user_preference(
        self,
        store_for: str,
        new_code: int,
        site_id: str,
        partition_id: str,
    ) -> Any:
        """Set user code preferences."""

        if store_for not in ["Arm", "Bypass"]:
            raise HyypApiError("Invalid selection, choose between Arm or Bypass")

        _params: dict[Any, Any] = STD_PARAMS.copy()
        _params["siteId"] = site_id

        _params["name"] = (
            "site." + site_id + ".partition." + partition_id + ".storeFor" + store_for
        )

        _params["preference_value"] = new_code

        try:
            req = self._session.post(
                "https://" + BASE_URL + API_ENDPOINT_SET_USER_PREFERANCE,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )

            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err

        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(
                f"Set user preferance failed with: {_json_result['error']}"
            )

        return _json_result

    def set_subuser_preference(
        self,
        user_id: str,
        site_id: str | None = None,
        partition_id: str | None = None,
        partition_pin: str | None = None,
        stay_profile_id: int | None = None,
    ) -> Any:
        """Set sub user preferences."""

        _params: dict[Any, Any] = STD_PARAMS.copy()
        _params["siteId"] = site_id
        _params["userId"] = user_id

        _params["partitions"] = {}
        _params["partitions"][0] = {}
        _params["partitions"][0][".id"] = partition_id
        _params["partitions"][0][".pin"] = partition_pin
        _params["stayProfileIds"] = {}
        _params["stayProfileIds"][0] = stay_profile_id

        try:
            req = self._session.post(
                "https://" + BASE_URL + API_ENDPOINT_UPDATE_SUB_USER,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )

            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err

        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(
                f"Updating sub user failed with: {_json_result['error']}"
            )

        return _json_result


    def pre_arm_check(
        self,        
        site_id: int,
        partition_id: int,
        pin: int | None = None,
        stay_profile_id: int | None = None
    ):
        
        
        if not pin:
            failure_info = {FAILURE_CAUSE_STRING : "NO PIN"}
            return failure_info        
        if not self.current_status:
            self.load_alarm_infos()
        zone_states = self.get_zone_state_info(site_id=site_id)       
        failed_zones = {}
        for zone_id_in_current_arm in self.current_status[site_id]["partitions"][partition_id]["zones"]:
            if zone_states is None:
                break
            if "zones" not in zone_states:
                break
            for zone_state in zone_states["zones"]:
                if self.current_status[site_id]["partitions"][partition_id]["zones"][zone_id_in_current_arm]['number'] != zone_state['number']:
                    continue
                if zone_state["bypassed"]:
                    break
                if not zone_state["openViolated"] and not zone_state["tampered"] :
                    break
                if stay_profile_id:
                    if zone_id_in_current_arm in self.current_status[site_id]["partitions"][partition_id]["stayProfiles"][stay_profile_id]["zoneIds"]:
                        break
                failed_zones[zone_state["number"]] = self.current_status[site_id]["partitions"][partition_id]["zones"][zone_id_in_current_arm]["name"]
                break
        if failed_zones:
            failure_info = {FAILURE_CAUSE_STRING : "VIOLATED ZONES",
                            "zones" : failed_zones
                            }
            return failure_info
        return {FAILURE_CAUSE_STRING : 0}
                
                        
    def arm_site(
        self,
        site_id: int,
        arm: bool = True,
        pin: int | None = None,
        partition_id: int | None = None,
        stay_profile_id: int | None = None,
    ) -> Any:
        """Arm alarm or stay profile via API."""

        _params: dict[Any, Any] = STD_PARAMS.copy()
        _params["arm"] = arm
        _params["pin"] = pin
        _params["partitionId"] = partition_id
        _params["siteId"] = site_id
        _params["stayProfileId"] = stay_profile_id
        del _params["imei"]
        _params["clientImei"] = STD_PARAMS["imei"]
        
        if arm:
            pre_check = self.pre_arm_check(site_id=site_id, partition_id=partition_id, pin=pin, stay_profile_id=stay_profile_id)
            pre_check = {"arm_fail_cause" : pre_check}
            
        try:
            req = self._session.get(
                "https://" + BASE_URL + API_ENDPOINT_ARM_SITE,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )

            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err
        
        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS":  
            if self.generic_callback_to_hass and arm:
                self.generic_callback_to_hass(data=pre_check)
            raise HyypApiError(f"Arm site failed: {_json_result['error']}")
            
        if self.generic_callback_to_hass:
            self.generic_callback_to_hass({"arm_fail_cause" : {FAILURE_CAUSE_STRING : 0}})
        return _json_result
        
        
        


    def trigger_alarm(
        self,
        site_id: int,
        pin: int | None = None,
        partition_id: int | None = None,
        trigger_id: int | None = None,
    ) -> Any:
        """Trigger Alarm via API."""

        _params: dict[Any, Any] = STD_PARAMS.copy()
        _params["pin"] = pin
        _params["partitionId"] = partition_id
        _params["siteId"] = site_id
        _params["triggerId"] = trigger_id
        del _params["imei"]
        _params["clientImei"] = STD_PARAMS["imei"]

        try:
            req = self._session.post(
                "https://" + BASE_URL + API_ENDPOINT_TRIGGER_ALARM,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )

            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err

        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(f"Trigger alarm failed: {_json_result['error']}")

        return _json_result


    def trigger_automation(
        self,
        site_id: int,
        trigger_id: int | None = None,
        pin: int | None = None,
    ) -> Any:
        """Trigger Automation via API."""

        _params: dict[Any, Any] = STD_PARAMS.copy()
        _params["pin"] = pin
        _params["siteId"] = site_id
        _params["triggerId"] = trigger_id
        del _params["imei"]
        _params["clientImei"] = STD_PARAMS["imei"]

        try:
            req = self._session.post(
                "https://" + BASE_URL + API_ENDPOINT_TRIGGER_AUTOMATION,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )

            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err

        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(f"Trigger automation failed: {_json_result['error']}")

        return _json_result


    def set_zone_bypass(
        self,
        zones: int,
        partition_id: int | None = None,
        stay_profile_id: int = 0,
        pin: int | None = None,
    ) -> Any:
        """Set/toggle zone bypass."""

        _params: dict[str, Any] = STD_PARAMS.copy()
        _params["partitionId"] = partition_id
        _params["zones"] = zones
        _params["stayProfileId"] = stay_profile_id
        _params["pin"] = pin
        del _params["imei"]
        _params["clientImei"] = STD_PARAMS["imei"]
        
        try:
            req = self._session.get(
                "https://" + BASE_URL + API_ENDPOINT_SET_ZONE_BYPASS,
                allow_redirects=False,
                params=_params,
                timeout=self._timeout,
            )
            req.raise_for_status()

        except requests.ConnectionError as err:
            raise InvalidURL("A Invalid URL or Proxy error occured") from err

        except requests.HTTPError as err:
            raise HTTPError from err

        try:
            _json_result: dict[Any, Any] = req.json()

        except ValueError as err:
            raise HyypApiError(
                "Impossible to decode response: "
                + str(err)
                + "\nResponse was: "
                + str(req.text)
            ) from err

        if _json_result["status"] != "SUCCESS" and _json_result["error"] is not None:
            raise HyypApiError(f"Failed to set zone bypass: {_json_result['error']}")

        return _json_result

    def logout(self) -> None:
        """Close ADT Secure Home session."""
        self.close_session()

    def close_session(self) -> None:
        """Clear current session."""
        if self._session:
            self._session.close()

        self._session = requests.session()
        self._session.headers.update(REQUEST_HEADER)  # Reset session.


"""Alarm info for hass integration."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any
from datetime import datetime
import time
from .constants import EventNumber, DEBUG_CLIENT_STRING
import logging
import threading as thread

_LOGGER = logging.getLogger(__name__)


if TYPE_CHECKING:
    from .client import HyypClient

SLEEP_DELAY = 0.5

class HyypAlarmInfos:
    """Initialize Hyyp alarm objects."""

    def __init__(self, client: HyypClient) -> None:
        """init."""
        self._client = client
        self._sync_info: dict = {}
        self._state_info: dict = {}
        self._notifications: dict = {}
        self._zone_state_info = []
        self._last_notification_check_timestamp = 0

    def _fetch_data(self, forced=False) -> None:
        """Fetch data via client api."""
        #forced = False
        if not forced:
            time.sleep(SLEEP_DELAY)
        self._sync_info = self._client.get_sync_info()
        if not forced:
            time.sleep(SLEEP_DELAY)
        self._state_info = self._client.get_state_info()
        if not forced:
            self._zone_state_info.clear()
            for site in self._sync_info["sites"]:
                siteid = site["id"]
                time.sleep(SLEEP_DELAY)
                site_zone_info = self._client.get_zone_state_info(site_id=siteid)
                entry = {siteid : site_zone_info}
                self._zone_state_info.append(entry)

            
        
    def get_zone_state_info_for_site(self, site):
        current_siteinfo = None
        for siteinfo in self._zone_state_info:
            if site in siteinfo:
                current_siteinfo = siteinfo[site]
                if current_siteinfo is None:
                    return None
                if "status" not in current_siteinfo:
                    return None
                if current_siteinfo["status"] != "SUCCESS":
                    return None
        return current_siteinfo
        
    def _fetch_notifications(self, site_id: int) -> dict[Any,Any]:
        """Fetch and cache site notifications."""
        self._notifications = self._client.site_notifications(site_id=site_id)


    def _last_notice(self) -> dict[Any, Any]:
        """Get last notification."""
        _response: dict[Any, Any] = {"lastNoticeTime": None, "lastNoticeName": None}

        if len(self._notifications) == 0:
            return _response
        
        _last_notification = self._notifications[0]
        
        if _last_notification:

            _last_event = _last_notification["eventNumber"]
            _last_event_datetime = str(
                datetime.fromtimestamp(_last_notification["timestamp"] / 1000)
            )  # Epoch in ms

            _response = {
                "lastNoticeTime": _last_event_datetime,
                "lastNoticeName": EventNumber[str(_last_event)],
            }

        return _response



    def _new_notifications(self) -> Any:

        _last_notification_check_timestamp = self._last_notification_check_timestamp   
        _response = []

        _notifications = self._notifications
                
        _current_timestamp = round(datetime.now().timestamp())
            
        for x in _notifications:
            
            _notification_timestamp = round(x['timestamp']/1000)
            if _current_timestamp - _notification_timestamp > 120:
                continue
            if _notification_timestamp <= (_last_notification_check_timestamp-30):
                continue
            _response.append(x)
        
        self._last_notification_check_timestamp = _current_timestamp
 
        return _response


    def _triggered_zones(self) -> Any:
           
        triggeredZoneIds = []
        _new_notifications = self._new_notifications()
        for _notification in _new_notifications:
            if _notification['eventNumber'] != 5:
                continue
            triggeredZoneIds.append(_notification['zoneId'])         
        _response = triggeredZoneIds
        
        return _response
     

     
    def _format_data(self) -> dict[Any, Any]:
        """Format data for Hass."""

        # The API returns data from site level.
        # Partitions are used as entity that actions are performed on.

        site_ids = {site["id"]: site for site in self._sync_info["sites"]}
        zone_ids = {zone["id"]: zone for zone in self._sync_info["zones"]}
        stay_ids = {
            stay_profile["id"]: stay_profile
            for stay_profile in self._sync_info["stayProfiles"]
        }
        partition_ids = {
            partition["id"]: partition for partition in self._sync_info["partitions"]
        }
        
        trigger_ids = {
            trigger["id"]: trigger for trigger in self._sync_info["triggers"]
        }
        
        for site in site_ids:
            
            self._fetch_notifications(site_id=site)
            triggered_zones = self._triggered_zones()
            zone_states = self.get_zone_state_info_for_site(site=site)

            # Add last site notification.
            _last_notice = self._last_notice()
            site_ids[site]["update"] = str(datetime.fromtimestamp(time.time()))
            site_ids[site]["lastNoticeTime"] = _last_notice["lastNoticeTime"]
            site_ids[site]["lastNoticeName"] = _last_notice["lastNoticeName"]

            # Add triggers (PGM / Automations in APP)
            site_ids[site]["triggers"] = {}
            for trigger_id in trigger_ids:
                if trigger_id not in site_ids[site]["triggerIds"]:
                    continue
                site_ids[site]["triggers"] = {
                    key: value
                    for (key, value) in trigger_ids.items()            
                }
                    
            # Add partition info.
            site_ids[site]["partitions"] = {
                partition_id: partition_ids[partition_id]
                for partition_id in site_ids[site]["partitionIds"]
            }

            for partition in partition_ids:
                # Add zone info to partition.
                if partition not in site_ids[site]["partitions"]:
                    continue
                site_ids[site]["partitions"][partition]["zones"] = {
                    key: value
                    for (key, value) in zone_ids.items()
                    if key in site_ids[site]["partitions"][partition]["zoneIds"]
                }

                # Add zone bypass info to zone.
                for zone in site_ids[site]["partitions"][partition]["zones"]:
                    site_ids[site]["partitions"][partition]["zones"][zone][
                        "bypassed"
                    ] = bool(zone in self._state_info["bypassedZoneIds"])

                # New zone information from IDS servers               
                for zone in site_ids[site]["partitions"][partition]["zones"]:
                    if zone_states is None:
                        continue
                    if "zones" not in zone_states:
                        continue
                    for zone_state in zone_states["zones"]:
                        if site_ids[site]["partitions"][partition]["zones"][zone][
                            "number"] != zone_state["number"]:
                            continue
                     #   site_ids[site]["partitions"][partition]["zones"][zone][
                     #       "bypassed"] = bool(zone_state["bypassed"] or 
                     #                          zone in self._state_info["bypassedZoneIds"])
                        site_ids[site]["partitions"][partition]["zones"][zone][
                            "openviolated"] = bool(zone_state["openViolated"])
                        site_ids[site]["partitions"][partition]["zones"][zone][
                            "tampered"] = bool(zone_state["tampered"])
                        site_ids[site]["partitions"][partition]["zones"][zone][
                            "stay_bypassed"] = False

                # Add zone trigger info to zone (Zone triggered alarm).
                for zone in site_ids[site]["partitions"][partition]["zones"]:
                    site_ids[site]["partitions"][partition]["zones"][zone][
                        "triggered"
                    ] = bool(zone in triggered_zones)


                # Add stay profile info.
                site_ids[site]["partitions"][partition]["stayProfiles"] = {
                    key: value
                    for (key, value) in stay_ids.items()
                    if key in site_ids[site]["partitions"][partition]["stayProfileIds"]
                }

                # Add partition armed status.
                site_ids[site]["partitions"][partition]["armed"] = bool(
                    partition in self._state_info["armedPartitionIds"]
                )

                # Add partition stay_armed status.
                site_ids[site]["partitions"][partition]["stayArmed"] = False
                
                for stay_profile in site_ids[site]["partitions"][partition]["stayProfiles"]:
                    if stay_profile not in self._state_info["armedStayProfileIds"]:
                        continue
                    site_ids[site]["partitions"][partition]["stayArmed"] = bool(
                        stay_profile in self._state_info["armedStayProfileIds"]
                    )
                    site_ids[site]["partitions"][partition]["stayArmedProfileName"] = (
                        site_ids[site]["partitions"][partition]["stayProfiles"][
                            stay_profile]["name"]
                    )
                    # Show zone as bypassed if stay partition has it bypassed                 
                    for zone in site_ids[site]["partitions"][partition]["zones"]:
                        bypassed_due_to_stay = zone in site_ids[site]["partitions"][partition]["stayProfiles"][stay_profile]['zoneIds']
                        site_ids[site]["partitions"][partition]["zones"][zone]["stay_bypassed"] = bypassed_due_to_stay
                        site_ids[site]["partitions"][partition]["zones"][zone][
                        "bypassed"
                        ] = bool(site_ids[site]["partitions"][partition]["zones"][zone]["bypassed"] or bypassed_due_to_stay)
                        
        return site_ids

    def status(self, forced=False) -> dict[Any, Any]:
        """Return the status of Hyyp connected alarms."""
        self._fetch_data(forced=forced)
        formatted_data: dict[Any, Any] = self._format_data()

        return formatted_data



    def debug_thread(self):
        while 1:
        
            _LOGGER.setLevel(logging.DEBUG)
            time.sleep(1.5)
            syncinfo = self._client.get_sync_info()
            time.sleep(1)
            stateinfo = self._client.get_state_info()
            time.sleep(2)
            notificationinfo = {}
            zoneinfo = {}
            for site in syncinfo["sites"]:  
                siteid = site["id"]
                time.sleep(2)
                cur_site_notificationinfo = self._client.site_notifications(site_id=siteid)
                notificationinfo[siteid] = cur_site_notificationinfo
                time.sleep(2)
                cur_site_zoneinfo = self._client.get_zone_state_info(site_id=siteid)
                zoneinfo[siteid] = cur_site_zoneinfo
                
            
            message = {"client_string" : DEBUG_CLIENT_STRING["client_string"],
                       "syncinfo" : syncinfo,
                       "Stateinfo" : stateinfo,
                       "notifications" : notificationinfo,
                       "zoneinfo" : zoneinfo,
                       }
            
            _LOGGER.debug('------------------------------')
            _LOGGER.debug('--------Start of debug--------')
            
            _LOGGER.debug(message)
            
            _LOGGER.debug('--------End of debug--------')
            _LOGGER.debug('-----------------------------')
            time.sleep(30)
    


    def get_debug_info(self) -> dict[Any, Any]:
        """Pull notifications for debug purposes."""
        # The API returns data from site level.
        thread.Thread(target=self.debug_thread).start()
        
    



import logging
import requests

_LOGGER = logging.getLogger(__name__)


class ClientTools:
    """Initialize api client object."""

    def __init__(
        self,
    ) -> None:
        """Initialize the client object."""
        
    def internet_connectivity(self):
        _LOGGER.debug("Checking for connectivity")
        try:
            reply = requests.get('http://www.msftconnecttest.com/connecttest.txt')
        except requests.exceptions.RequestException:
            return False
        if reply.status_code != 200:
            return False
        if reply.text != "Microsoft Connect Test":
            return False
        return True